/** GStundiscret
 *
 * A simple nagravideo decrypt plugin for gstreamer0.10
 *
 *   Copyright (C) 2006  megabibi@gmail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 */


/**
 * SECTION:element-plugin
 *
 * <refsect2>
 * <title>Example launch line</title>
 * <para>
 * <programlisting>
 * gst-launch v4l2src ! undiscret silent=true ! xvimagesink
 * </programlisting>
 * </para>
 * </refsect2>
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include "gstundiscret.h"

//not good, two times the same argument in the macro
#define SQUARE(x) ((x)*(x))
#define min(X, Y)  ((X) < (Y) ? (X) : (Y))
#include <math.h>

/* Filter signals and args */
enum {
  /* FILL ME */
  LAST_SIGNAL
};

enum {
  ARG_0,
  ARG_SILENT
};


///TODO remove the limitation on the widt/height and detect them when the negociation of the caps is done
static GstStaticPadTemplate sink_factory =
GST_STATIC_PAD_TEMPLATE (
  "sink",
  GST_PAD_SINK,
  GST_PAD_ALWAYS,
  GST_STATIC_CAPS ("video/x-raw-yuv, "
		   "width=(int)720, "
		   "height=(int)576, "
		   "format=(fourcc){YUY2}"
)
);

static GstStaticPadTemplate src_factory =
GST_STATIC_PAD_TEMPLATE (
  "src",
  GST_PAD_SRC,
  GST_PAD_ALWAYS,
  GST_STATIC_CAPS ("video/x-raw-yuv, "
		"width=(int)720, "
		"height=(int)576, "
		   "format=(fourcc){YUY2}"
)
);

static void	gst_undiscret_class_init	 (GstundiscretClass *klass);
static void	gst_undiscret_base_init	 (GstundiscretClass *klass);
static void	gst_undiscret_init	 (Gstundiscret *filter,
                                                  GstundiscretClass *gclass);

static void	gst_undiscret_set_property (GObject *object, guint prop_id,
                                                  const GValue *value,
					          GParamSpec *pspec);
static void	gst_undiscret_get_property (GObject *object, guint prop_id,
                                                  GValue *value,
						  GParamSpec *pspec);

static gboolean gst_undiscret_set_caps (GstPad *pad, GstCaps *caps);
static GstFlowReturn gst_undiscret_chain (GstPad *pad, GstBuffer *buf);

static GstElementClass *parent_class = NULL;

GType
gst_undiscret_get_type (void)
{
  static GType plugin_type = 0;

  if (!plugin_type)
  {
    static const GTypeInfo plugin_info =
    {
      sizeof (GstundiscretClass),
      (GBaseInitFunc) gst_undiscret_base_init,
      NULL,
      (GClassInitFunc) gst_undiscret_class_init,
      NULL,
      NULL,
      sizeof (Gstundiscret),
      0,
      (GInstanceInitFunc) gst_undiscret_init,
    };
    plugin_type = g_type_register_static (GST_TYPE_ELEMENT,
	                                  "Gstundiscret",
	                                  &plugin_info, 0);
  }
  return plugin_type;
}

static void
gst_undiscret_base_init (GstundiscretClass *klass)
{
  static GstElementDetails element_details = {
    "unDiscret Plugin",
    "Filter/Effect/Video",
    "A video unDiscret plugin",
    "Bibi"
  };
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);

  gst_element_class_add_pad_template (element_class,
	gst_static_pad_template_get (&src_factory));
  gst_element_class_add_pad_template (element_class,
	gst_static_pad_template_get (&sink_factory));
  gst_element_class_set_details (element_class, &element_details);
}

/* initialize the plugin's class */
static void
gst_undiscret_class_init (GstundiscretClass *klass)
{
  GObjectClass *gobject_class;
  GstElementClass *gstelement_class;

  gobject_class = (GObjectClass*) klass;
  gstelement_class = (GstElementClass*) klass;

  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = gst_undiscret_set_property;
  gobject_class->get_property = gst_undiscret_get_property;
  
  g_object_class_install_property (gobject_class, ARG_SILENT,
    g_param_spec_boolean ("silent", "Silent", "Produce verbose output ?",
                          FALSE, G_PARAM_READWRITE));
}



static struct YUY2twopixels rgb2yuv(int R,int G,int B)
{
  struct YUY2twopixels res;
  res.Y0=(guint8) ( (  66 * R + 129 * G +  25 * B + 128) >> 8) +  16;
  res.Y1=(guint8) ( (  66 * R + 129 * G +  25 * B + 128) >> 8) +  16;
  res.U0=(guint8) ( ( -38 * R -  74 * G + 112 * B + 128) >> 8) + 128;
  res.V0=(guint8) ( ( 112 * R -  94 * G -  18 * B + 128) >> 8) + 128;
  return res;
}


/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
gst_undiscret_init (Gstundiscret *filter,
    GstundiscretClass * gclass)
{
  GstElementClass *klass = GST_ELEMENT_GET_CLASS (filter);

  filter->sinkpad = gst_pad_new_from_template (
	gst_element_class_get_pad_template (klass, "sink"), "sink");
  gst_pad_set_setcaps_function (filter->sinkpad, gst_undiscret_set_caps);
  gst_pad_set_getcaps_function (filter->sinkpad, gst_pad_proxy_getcaps);

  filter->srcpad = gst_pad_new_from_template (
	gst_element_class_get_pad_template (klass, "src"), "src");
  gst_pad_set_getcaps_function (filter->srcpad, gst_pad_proxy_getcaps);

  gst_element_add_pad (GST_ELEMENT (filter), filter->sinkpad);
  gst_element_add_pad (GST_ELEMENT (filter), filter->srcpad);
  gst_pad_set_chain_function (filter->sinkpad, gst_undiscret_chain);



  filter->silent = FALSE;
  ///TODO make percent a parameter
  filter->percent= 0.1;
  filter->thr = 50;
  filter->rightmargin = 10;

  ///TODO How can we get these values ?
  filter->width= 720;
  filter->height=576;


  filter->n_blck_pixel = floor(filter->width*(902e-3/52));
  if (filter->percent==0)
	filter->step= (filter->width-6*filter->n_blck_pixel)*2;
  else
	filter->step= min((filter->width-6*filter->n_blck_pixel)*2 , 1.0/filter->percent);


  filter->depl2=malloc(sizeof(int)*(filter->height));
  filter->depl=malloc(sizeof(int)*(filter->height));
  filter->depl_confidence=malloc(sizeof(int)*(filter->height));
  filter->depl2cf=malloc(sizeof(int)*(filter->height));



  /*Setting the elements red,green,blue*/
  filter->red=rgb2yuv(255,0,0);
  filter->green=rgb2yuv(0,255,0);
  filter->blue=rgb2yuv(0,0,255);
}

static void
gst_undiscret_set_property (GObject *object, guint prop_id,
                                  const GValue *value, GParamSpec *pspec)
{
  Gstundiscret *filter = GST_UNDISCRET (object);

  switch (prop_id)
  {
  case ARG_SILENT:
    filter->silent = g_value_get_boolean (value);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

static void
gst_undiscret_get_property (GObject *object, guint prop_id,
                                  GValue *value, GParamSpec *pspec)
{
  Gstundiscret *filter = GST_UNDISCRET (object);

  switch (prop_id) {
  case ARG_SILENT:
    g_value_set_boolean (value, filter->silent);
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    break;
  }
}

/* GstElement vmethod implementations */

/* this function handles the link with other elements */
static gboolean
gst_undiscret_set_caps (GstPad *pad, GstCaps *caps)
{
  Gstundiscret *filter;
  GstPad *otherpad;

  filter = GST_UNDISCRET (gst_pad_get_parent (pad));
  otherpad = (pad == filter->srcpad) ? filter->sinkpad : filter->srcpad;

  return gst_pad_set_caps (pad, caps);
}


/*
  return 0, 1 or 2 if we are able to guess the shift only looking at
  the left border of the image. Should work on not too dark image

  return -1 if we are not able to guess*/
static int left_guess(Gstundiscret *filter, unsigned char * image, int lineNb)
{
  int n_blck_pixel = filter->n_blck_pixel;
  int thr= filter->thr;
  int var = 3;//we don't consider the "var" pixeles around the estimated boundary between the black and the image

  int i;
  //printf("entering left_guess of line : %i\n",lineNb);
  for (i=lineNb * filter->width*2; i< (lineNb * filter->width+n_blck_pixel-var)*2  ; i+=2)
    if (image[i] >= thr)
      return 0;
  int allless, allup;
  allless=1;
  allup=1;

  for(i=(lineNb * filter->width+n_blck_pixel+var)*2 ; i< (lineNb * filter->width+2*n_blck_pixel-var)*2; i+=2)
    if (image[i] >= thr)
      allless=0;
    else
      allup=0;

  if (allup)
    return 1;
  else
    if (!allless)
      return -1;

  allless=1;
  allup=1;

  for(i=(lineNb * filter->width+2*n_blck_pixel+var)*2 ; i< (lineNb * filter->width+3*n_blck_pixel-var)*2; i+=2)
    if (image[i] >= thr)
      allless=0;
    else
      allup=0;
  
  if (allup)
    return 2;
  else
    if (!allless)
      return -1;

  return -1;
}

/* return an estimation of the distance between two lines of an image
   if we make a shift of n x n_blck_pixel of the second line
   with n in [-3 3] */
static long distance(Gstundiscret *filter, unsigned char * image, int line1, int line2,int n)
{
  ///TODO take into account the value percent.
  int width=filter->width;
  int n_blck_pixel=filter->n_blck_pixel;
  int l1start= (line1*width+3*n_blck_pixel)*2;
  int l2start= (line2*width+(3+n)*n_blck_pixel)*2;
  long dis=0;
  int i;
  //printf("step %i sur %i\n",filter->step,filter->width);
  for (i=0; i < (width-6*n_blck_pixel)*2; i+=filter->step)
    dis+= SQUARE((long) image[l1start+i] - (long) image[l2start+i] );
  if (dis<0) printf("BIG error, distance negative");
  return dis;
}

static int *  makeFeasable(Gstundiscret * filter, int * depl,int * conf){

	int * depl2 = filter->depl2;// malloc(sizeof(int)*(filter->height));
	int * depl2cf =filter->depl2cf;// malloc(sizeof(int)*(filter->height));

	int i,j,k;

	j=0;
	while ((conf[j]==0) || ((depl[j]!=-2)&& (depl[j]!=2) ) && (j < filter->height-1) )
		j++;

	//j must now correspond to a sure -2 or +2 shift or the end the picture height-1 (black picture ?)
	if (depl[j]==-2)
		depl2[j]=0;
	else
		depl2[j]=2;
	depl2cf[j]=1;

	for(i=j-1 ; i>=0; --i){
		depl2[i]=depl2[i+1]-depl[i+1];
		if (depl2[ i ] < 0) {depl2[ i ] = 0;}
		if (depl2[ i ] > 2) {depl2[ i ] = 2;}
	}

	//Complete the rest

	//simple (working) completion
// 	for(i=j+1 ; i < filter->height ; i++){
// 		depl2[ i ] = depl2[ i-1 ] + depl[ i ];
// 		if (depl2[ i ] < 0) {depl2[ i ] = 0;}
// 		if (depl2[ i ] > 2) {depl2[ i ] = 2;}
// 	}

	//complex completion (for text generic & black image)
	i=j+1;
	while (i < filter->height ){
		if (((depl[i]==-2) || (depl[i]==2)) && (conf[i]==1) ){ //we are sure
			if (depl[i]==-2)
				depl2[i]=0;
			else
				depl2[i]=2;
			depl2cf[i]=1;

			//Go back up to take into account this sure deplacement
			k=i-1;
			while (depl2cf[k] ==0){ //k should not decrease less than j
				depl2[k]=depl2[k+1]-depl[k+1];
				if (depl2[ k ] < 0) {depl2[ k ] = 0;}
				if (depl2[ k ] > 2) {depl2[ k ] = 2;}
				--k;
			}
		}
		else{  //simply complete
			depl2[ i ] = depl2[ i-1 ] + depl[ i ];
			if (depl2[ i ] < 0) {depl2[ i ] = 0;}
			if (depl2[ i ] > 2) {depl2[ i ] = 2;}
			//is this value sure ?
			if ((depl2cf[i-1]==1) && (conf[i]==1))
				depl2cf[i]=1;
			else
				depl2cf[i]=0;
		}
		++i;
	}


	//free(depl2cf);

	return depl2;
}







/* chain function
 * this function does the actual processing
 */

static GstFlowReturn
gst_undiscret_chain (GstPad *pad, GstBuffer *buf)
{
  Gstundiscret *filter;

  filter = GST_UNDISCRET (GST_OBJECT_PARENT (pad));
  GstBuffer * out_buf;

  out_buf = gst_buffer_make_writable (buf);
//   struct YUY2twopixels  * out_data = (struct YUY2twopixels *) (GST_BUFFER_DATA(out_buf));
//   struct YUY2twopixels  * in_data  = (struct YUY2twopixels *) (GST_BUFFER_DATA(buf));

   //two pointers at the begining of the datas (2 bytes per pixel)
   unsigned char * g_out = (unsigned char *) (GST_BUFFER_DATA(out_buf));
   unsigned char * g_in = (unsigned char *) (GST_BUFFER_DATA(out_buf));

  // getchar();
  //printf("distance between line 10 and line 20 : %d\n",distance(filter,(unsigned char *) in_data,10,20,0));
  //printf("distance between line 10 and line 500 :  %d\n",distance(filter,(unsigned char *) in_data,10,500,0));


	int i,j;
	long k;
	int shift;
	int * depl = filter->depl; //malloc(sizeof(int)*(filter->height));
	int * depl_confidence = filter->depl_confidence; //malloc(sizeof(int)*(filter->height));
	depl[0]=-1; //we don't know the shift
	depl_confidence[0]=0;

	long dis[5]; //contains the distances corresponding to a shift -2 -1 0 1 2
	for(i=0 ; i < filter->height-1 ; ++i){
		for (j=-2 ; j<3 ; ++j)
			dis[j+2]=distance(filter, g_in, i,i+1,j);
		long min1,min2,shift;
		min1=dis[0];
		shift=-2;
		for(j=1; j<5;++j)
			if (dis[j]<min1){min1=dis[j];shift=j-2;}
		min2=2000000000;
		for(j=0; j<5;++j)
			if ((dis[j]<min2) && (j-2!=shift) ){min2=dis[j];}
		depl[i+1]=shift;
		//printf("relative move of line %i of %i since %ld < %ld : %f \n",i,depl[i],min1,min2,(double)min1/(double)min2);

		if ((double)min1/(double)min2 < 0.15) //the prediction must be good
			depl_confidence[i+1]=1;
		else
			depl_confidence[i+1]=0;
	}



  ///TODO implement a function to detect the shift in function of the first pixels of each line

	//a list of the deplacements

	int * depl2 = makeFeasable(filter,depl,depl_confidence);

	//free(depl_confidence);

	//apply the displacements of depl2
	for(i=0 ; i < filter->height ; ++i){
		//printf("moving line %i of %i\n",i,depl2[i]);

		//printf("move de la ligne %i de %i * %i pixels\n",i,depl2[i],filter->n_blck_pixel);
		for(j=0  ;  j< filter->width - depl2[i]*filter->n_blck_pixel - filter->rightmargin ; ++j){
			g_out[(i*filter->width+j)*2  ]=g_in[(i*filter->width+j+depl2[i]*filter->n_blck_pixel)*2 ];
			g_out[(i*filter->width+j)*2+1]=g_in[(i*filter->width+j+depl2[i]*filter->n_blck_pixel)*2+1];
		}
		//     for(j = filter->width - depl2[i]*filter->n_blck_pixel -filter->rightmargin ; j < filter->width ;++j){
		//       g_out[(i*filter->width+j)*2  ]=g_in[(i*filter->width+j)*2 ];
		//       g_out[(i*filter->width+j)*2+1]=g_in[(i*filter->width+j)*2+1];
		//     }
		
		//treatment of the right border

		int even =((filter->width)%2 == 0);

		for(j = filter->width - depl2[i]*filter->n_blck_pixel -filter->rightmargin ; j < filter->width ;++j){
			if ((even && (j%2==0))||(!even && !(j%2==0))) {
			g_out[(i*filter->width+j)*2  ]=g_in[((i+1)*filter->width-filter->rightmargin )*2 -4 ];
			g_out[(i*filter->width+j)*2+1]=g_in[((i+1)*filter->width-filter->rightmargin)*2 -3 ];
			}
			else{
			g_out[(i*filter->width+j)*2  ]=g_in[((i+1)*filter->width-filter->rightmargin)*2 -2 ];
			g_out[(i*filter->width+j)*2+1]=g_in[((i+1)*filter->width-filter->rightmargin)*2 -1 ];
			}
		}
	}



    //free(depl);
    //free(depl2);


/*
for(i=0 ; i < filter->height ; ++i)
  for(j=0  ;  j< filter->width  ; ++j)
    g_out[(i*filter->width+j)*2  ]=0;
printf("width=%i,height=%i\n",filter->width,filter->height);*/


   //Code for debugging
  //out_data = (struct YUY2twopixels *) (GST_BUFFER_DATA(out_buf));

/*   for(i=200;i<240 ; ++i){ */
/*     for (j=0 ; j<60 ; ++j) */
/*       printf("%i ", g[(i*filter->width+j)*2]); */
/*     printf("\n"); */
/*     for (j=0 ; j<60 ; ++j) */
/*       printf("%i ", g[(i*filter->width+j)*2+1]); */
/*     printf("\n"); */


/*     j=left_guess(filter,(unsigned char *) in_data,i); */
/*     printf("guess of line %i : %i\n",i,j); */
/*     if (j==0){ */
/*       out_data[(i*filter->width+80)/2]=filter->red; */
/*       out_data[(i*filter->width+82)/2]=filter->red; */
/*       out_data[(i*filter->width+84)/2]=filter->red; */
/*     } */
/*     if (j==1){ */
/*       out_data[(i*filter->width+80)/2]=filter->green; */
/*       out_data[(i*filter->width+82)/2]=filter->green; */
/*       out_data[(i*filter->width+84)/2]=filter->green; */
/*     } */
/*     if (j==2){ */
/*       out_data[(i*filter->width+80)/2]=filter->blue; */
/*       out_data[(i*filter->width+82)/2]=filter->blue; */
/*       out_data[(i*filter->width+84)/2]=filter->blue; */
/*     } */
      
        
/*     printf("\n\n"); */
/*   } */


  if (filter->silent == FALSE){
    //printf("width=%i,height=%i\n",filter->width,filter->height);
    //printf("nb_blck_block=%i",filter->n_blck_pixel);
  }
 

  return gst_pad_push (filter->srcpad, out_buf);
}


/* entry point to initialize the plug-in
 * initialize the plug-in itself
 * register the element factories and pad templates
 * register the features
 *
 * exchange the string 'plugin' with your elemnt name
 */
static gboolean
plugin_init (GstPlugin *plugin)
{
  return gst_element_register (plugin, "undiscret",
			       GST_RANK_NONE,
			       GST_TYPE_UNDISCRET);
}

/* this is the structure that gstreamer looks for to register plugins
 *
 * exchange the strings 'plugin' and 'Template plugin' with you plugin name and
 * description
 */
GST_PLUGIN_DEFINE (
  GST_VERSION_MAJOR,
  GST_VERSION_MINOR,
  "undiscret",
  "canal+ image decrypt plugin",
  plugin_init,
  VERSION,
  "LGPL",
  "GStreamer",
  "http://gstreamer.net/"
)
