/** soundcp.c
 *
 * A simple nagraudio decrypt program
 *
 * Use "gcc -o soundcp `pkg-config --cflags --libs jack` soundcp.c" to compile
 *
 *   Copyright (C) 2006  megabibi@gmail.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Some parts taken from the Chorus plugin Copyright (c) 2002 Grincheux
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include "math.h"
#include <jack/jack.h>
#include <sys/ioctl.h>
#include <linux/videodev.h>
#include <fcntl.h>
#include <signal.h>


char * devicename="/dev/video0";

void open_sound_output()
{
  int fd = open(devicename, O_RDWR, 0);
  struct v4l2_control c;
  c.id= V4L2_CID_AUDIO_MUTE ;
  c.value = 0;
  if(ioctl(fd, VIDIOC_S_CTRL, &c) == -1) {
        char s[80];
        sprintf(s, "VIDIOC_S_CTRL, id=%d, value=%d", c.id, c.value);
        perror(s);
  }
}

void close_sound_output()
{
  int fd = open(devicename, O_RDWR, 0);
  struct v4l2_control c;
  c.id= V4L2_CID_AUDIO_MUTE ;
  c.value = 1;
  if(ioctl(fd, VIDIOC_S_CTRL, &c) == -1) {
        char s[80];
        sprintf(s, "VIDIOC_S_CTRL, id=%d, value=%d", c.id, c.value);
        perror(s);
  }
}


/*Definition of the global variables for the sound processing*/
jack_port_t *input_port;
jack_port_t *output_port;
jack_client_t *client;

int rate;
int pitch;
int i;
double m_fPostXV[7+1];
double m_fPostYV[7+1];
double m_fPreXV[4+1];
double m_fPreYV[4+1];
double w, m_dDeccte, m_dDecy1, m_dDecy0 ;
double coeff_pre[5];
double coeff_post[8];
int m_nPitch;
int m_nPrefilter=1;
int m_nPostfilter=1;
int m_nGain = 1;
int boost;
double b = M_PI / 2;
double decy;

static void
ZeroMemory( double * tableau, int n)
{
  double * tab = (double *) tableau;
  int i;
  for (i=0 ; i<n ; ++i)
    tab[i] = 0;
} 


void initFilterParams(int soundrate)
{
	//Init the variable for the sound processing
	rate=soundrate;
	pitch = 0;
	
	ZeroMemory( m_fPostXV, sizeof(m_fPostXV));
	ZeroMemory( m_fPostYV, sizeof(m_fPostYV));
	ZeroMemory( m_fPreXV, sizeof(m_fPreXV));
	ZeroMemory( m_fPreYV, sizeof(m_fPreYV));
	
	boost=(m_nPrefilter == 2)?3:2;
	w = 2 * M_PI * 12800 / ((double) rate + m_nPitch);
	m_dDeccte = 2.0 * cos(w);
	m_dDecy0 = sin(-2 * w + b);
	m_dDecy1 = sin(-w + b);
	m_nPitch=pitch;
	
	//Init prefilter
	if (rate < 38000)
	{
		coeff_pre[0]=1.497238168;
		coeff_pre[1]=-0.4498778345;
		coeff_pre[2]=-2.1450223211;
		coeff_pre[3]=-3.8977979094;
		coeff_pre[4]=-3.1936445246;
	
		coeff_post[0]=1.058175797e+01;
		coeff_post[1]=0.0158398918;
		coeff_post[2]=-0.0989743038;
		coeff_post[3]=-0.4806770487;
		coeff_post[4]=-1.5976153081;
		coeff_post[5]=-2.8744063538;
		coeff_post[6]=-3.5609112971;
		coeff_post[7]=-2.4995443421;
		boost=2;//don't work with low Frequency
	}
	else if (rate < 46000)
	{
		coeff_pre[0]=6.310833714;
		coeff_pre[1]=-0.0622354481;
		coeff_pre[2]=-0.0826699057;
		coeff_pre[3]=-0.7552376542;
		coeff_pre[4]=-0.6351796968;
	
		coeff_post[0]=2.211596484e+02;
		coeff_post[1]=0.0935164959;
		coeff_post[2]=-0.4657366789;
		coeff_post[3]=1.2531881677;
		coeff_post[4]=-2.2854956347;
		coeff_post[5]=2.9413622606;
		coeff_post[6]=-2.9346526015;
		coeff_post[7]=1.8190505669;
	
		boost=2;//don't work with low Frequency
	}
	else if (rate < 72000)
	{
		coeff_pre[0]=8.183340972;
		coeff_pre[1]=-0.0626606278;
		coeff_pre[2]=0.0379954621;
		coeff_pre[3]=-0.6759257285;
		coeff_pre[4]=-0.2546007635;
	
		coeff_post[0]=3.834721196e+02;
		coeff_post[1]=0.1130488058;
		coeff_post[2]=-0.6208360469;
		coeff_post[3]=1.7554209098;
		coeff_post[4]=-3.2416570827;
		coeff_post[5]=4.1799016509;
		coeff_post[6]=-3.9042013855;
		coeff_post[7]=2.3845309549;
	}
	else
	{
		coeff_pre[0]=7.486320862e+01;
		coeff_pre[1]=-0.2082551105;
		coeff_pre[2]=0.9768363336;
		coeff_pre[3]=-1.9591638847;
		coeff_pre[4]=1.9768595219;
	
		coeff_post[0]=3.508040262e+04;
		coeff_post[1]=0.3339406563;
		coeff_post[2]=-2.4120965667;
		coeff_post[3]=7.7656156821;
		coeff_post[4]=-14.4501981950,
		coeff_post[5]=16.8069704710;
		coeff_post[6]=-12.2491585080;
		coeff_post[7]=5.2012776995;
	}

}


/**
 * The process callback for this JACK application is called in a
 * special realtime thread once for each audio cycle.
 *
 * This client follows a simple rule: when the JACK transport is
 * running, copy the input port to the output.  When it stops, exit.
 */
int
process (jack_nframes_t nframes, void *arg)
{
	jack_default_audio_sample_t *in, *out;
	jack_transport_state_t ts = jack_transport_query(client, NULL);

	//if (ts == JackTransportRolling) {

		in = jack_port_get_buffer (input_port, nframes);
		out = jack_port_get_buffer (output_port, nframes);

		//processing the sound
		double data_in,data_out;
		for ( i = 0 ; i < nframes; i ++ ){
			data_in=in[i];
			if ( m_nPrefilter )
			{
			m_fPreXV[0] = m_fPreXV[1]; m_fPreXV[1] = m_fPreXV[2]; m_fPreXV[2] = m_fPreXV[3];
			m_fPreXV[3] = m_fPreXV[4]; 
			m_fPreXV[4] =(double)((double) data_in /(coeff_pre[0]));//GDS
						
			m_fPreYV[0] = m_fPreYV[1]; m_fPreYV[1] = m_fPreYV[2]; m_fPreYV[2] = m_fPreYV[3];
			m_fPreYV[3] = m_fPreYV[4];
			m_fPreYV[4] =(double)  ((m_fPreXV[0] + m_fPreXV[4])
						+ 4 * (m_fPreXV[1] + m_fPreXV[3]) + 6 * m_fPreXV[boost]
						+ ( coeff_pre[1] * m_fPreYV[0]) + ( coeff_pre[2] * m_fPreYV[1])
						+ ( coeff_pre[3] * m_fPreYV[2]) + ( coeff_pre[4] * m_fPreYV[3]));
						
			data_in = m_fPreYV[4];
			}
				
			decy=(m_dDeccte * m_dDecy1) - m_dDecy0;  
			m_dDecy0=m_dDecy1;
			m_dDecy1=decy;
			data_out =  (data_in * decy);
		
			if ( m_nPostfilter)
			{
			m_fPostXV[0] = m_fPostXV[1]; m_fPostXV[1] = m_fPostXV[2]; m_fPostXV[2] = m_fPostXV[3];
			m_fPostXV[3] = m_fPostXV[4]; m_fPostXV[4] = m_fPostXV[5]; m_fPostXV[5] = m_fPostXV[6];
			m_fPostXV[6] = m_fPostXV[7]; 
						
			m_fPostXV[7] = (double)( (double)(data_out) / coeff_post[0] );
						
			m_fPostYV[0] = m_fPostYV[1]; m_fPostYV[1] = m_fPostYV[2]; m_fPostYV[2] = m_fPostYV[3];
			m_fPostYV[3] = m_fPostYV[4]; m_fPostYV[4] = m_fPostYV[5]; m_fPostYV[5] = m_fPostYV[6];
			m_fPostYV[6] = m_fPostYV[7]; 
						
			m_fPostYV[7] =  (double) ((m_fPostXV[0] + m_fPostXV[7])
						+ 7 * (m_fPostXV[1] + m_fPostXV[6]) + 21 * (m_fPostXV[2] + m_fPostXV[5])
						+ 35 * (m_fPostXV[3] + m_fPostXV[4]) 
						+ ( coeff_post[1] * m_fPostYV[0]) + (  coeff_post[2] * m_fPostYV[1])
						+ ( coeff_post[3] * m_fPostYV[2]) + (  coeff_post[4] * m_fPostYV[3])
						+ ( coeff_post[5] * m_fPostYV[4]) + (  coeff_post[6] * m_fPostYV[5])
						+ ( coeff_post[7] * m_fPostYV[6]) );
			data_out=m_fPostYV[7];
			}
			data_out = data_out * m_nGain;
		
			out[i]=data_out;
		}
	//}

	return 0;
}

/**
 * JACK calls this shutdown_callback if the server ever shuts down or
 * decides to disconnect the client.
 */
void
jack_shutdown (void *arg)
{
	close_sound_output();
	exit (1);
}


/* To cut the sound off when user hit C-c */
void catch_int(int sig_num)
{
	jack_client_close (client);
	close_sound_output();
	exit(0);
}


int
main (int argc, char *argv[])
{
	signal(SIGINT, catch_int);
	//open the sound of the tv card
	//TODO to parametrize the value of "/dev/video0".
	open_sound_output();

	int defRate=48000;
	initFilterParams(defRate);  //TODO default rate should be parametrize or find where we can get the actual rate and initialize then
	printf("filter rate set to %i Hz\n",defRate);

	const char **ports;
	const char *client_name;
	const char *server_name = NULL;
	jack_options_t options = JackNullOption;
	jack_status_t status;



	if (argc >= 2) {		/* client name specified? */
		client_name = argv[1];
		if (argc >= 3) {	/* server name specified? */
			server_name = argv[2];
			options |= JackServerName;
		}
	} else {			/* use basename of argv[0] */
		client_name = strrchr(argv[0], '/');
		if (client_name == 0) {
			client_name = argv[0];
		} else {
			client_name++;
		}
	}


	/* open a client connection to the JACK server */
	client = jack_client_open (client_name, options, &status, server_name);
	if (client == NULL) {
		fprintf (stderr, "jack_client_open() failed, "
			"status = 0x%2.0x\n", status);
		if (status & JackServerFailed) {
			fprintf (stderr, "Unable to connect to JACK server\n");
		}
		exit (1);
	}
	if (status & JackServerStarted) {
		fprintf (stderr, "JACK server started\n");
	}
	if (status & JackNameNotUnique) {
		client_name = jack_get_client_name(client);
		fprintf (stderr, "unique name `%s' assigned\n", client_name);
	}


	/* tell the JACK server to call `process()' whenever
	there is work to be done.
	*/

	jack_set_process_callback (client, process, 0);

	/* tell the JACK server to call `jack_shutdown()' if
	it ever shuts down, either entirely, or if it
	just decides to stop calling us.
	*/

	jack_on_shutdown (client, jack_shutdown, 0);

	/* display the current sample rate. 
	*/
	int temprate = (int) jack_get_sample_rate (client);
	printf ("engine sample rate: %i Hz\nShould be the same as the filter rate\n",
		temprate);

	/* create two ports */
	input_port = jack_port_register (client, "input",
					JACK_DEFAULT_AUDIO_TYPE,
					JackPortIsInput, 0);
	output_port = jack_port_register (client, "output",
					JACK_DEFAULT_AUDIO_TYPE,
					JackPortIsOutput, 0);

	if ((input_port == NULL) || (output_port == NULL)) {
		fprintf(stderr, "no more JACK ports available\n");
		exit (1);
	}

	/* Tell the JACK server that we are ready to roll.  Our
	* process() callback will start running now. */

	if (jack_activate (client)) {
		fprintf (stderr, "cannot activate client");
		exit (1);
	}

	/* Connect the ports.  You can't do this before the client is
	* activated, because we can't make connections to clients
	* that aren't running.  Note the confusing (but necessary)
	* orientation of the driver backend ports: playback ports are
	* "input" to the backend, and capture ports are "output" from
	* it.
	*/

	ports = jack_get_ports (client, NULL, NULL,
				JackPortIsPhysical|JackPortIsOutput);
	if (ports == NULL) {
		fprintf(stderr, "no physical capture ports\n");
		exit (1);
	}

	if (jack_connect (client, ports[0], jack_port_name (input_port))) {
		fprintf (stderr, "cannot connect input ports\n");
	}

	free (ports);
	
	ports = jack_get_ports (client, NULL, NULL,
				JackPortIsPhysical|JackPortIsInput);
	if (ports == NULL) {
		fprintf(stderr, "no physical playback ports\n");
		exit (1);
	}

	if (jack_connect (client, jack_port_name (output_port), ports[0])) {
		fprintf (stderr, "cannot connect output ports\n");
	}

	free (ports);

	/* keep running until the transport stops */
	for(;;)
	sleep (1);

	jack_client_close (client);
	close_sound_output();
	exit (0);
}
